# Copyright (c) 2023 Evan Palmer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""An enum wrapper that provides an abstraction for the ID of a Packet.

The PacketID provides an abstraction for each of the packet IDs supported by the Reach
serial protocol.

Examples:
    >>> PacketID.POSITION.value
    0x03
    >>> PacketID(1)
    PacketID.MODE
"""

from enum import Enum


class PacketID(Enum):
    """The ID of a packet."""

    MODE = 0x01
    VELOCITY = 0x02
    POSITION = 0x03
    CURRENT = 0x05
    RELATIVE_POSITION = 0x0E
    INDEXED_POSITION = 0x0D
    REQUEST = 0x60
    SERIAL_NUMBER = 0x61
    MODEL_NUMBER = 0x62
    TEMPERATURE = 0x66
    SOFTWARE_VERSION = 0x6C
    KM_END_POS = 0xA1
    KM_END_VEL = 0xA2
    KM_END_VEL_LOCAL = 0xCB
    KM_BOX_OBSTACLE_02 = 0xA5
    KM_BOX_OBSTACLE_03 = 0xA6
    KM_BOX_OBSTACLE_04 = 0xA7
    KM_BOX_OBSTACLE_05 = 0xA8
    KM_CYLINDER_OBSTACLE_02 = 0xAB
    KM_CYLINDER_OBSTACLE_03 = 0xAC
    KM_CYLINDER_OBSTACLE_04 = 0xAD
    KM_CYLINDER_OBSTACLE_05 = 0xAE
    VOLTAGE = 0x90
    SAVE = 0x50
    HEARTBEAT_FREQUENCY = 0x92
    HEARTBEAT_SET = 0x91
    POSITION_LIMITS = 0x10
    VELOCITY_LIMITS = 0x11
    CURRENT_LIMITS = 0x12
    ATI_FT_READING = 0xD8
    BOOTLOADER = 0xFF
    VOLTAGE_THRESHOLD_PARAMETERS = 0x99



# Copyright (c) 2023 Evan Palmer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

r"""Provides a wrapper for packets defined using Reach serial protocol.

A ``Packet`` provides an abstraction for serial data received from the Reach Bravo 7
manipulator. A ``Packet`` can also be used to create commands to send to the Bravo.

Examples:
    >>> Packet(DeviceID.LINEAR_JAWS, PacketID.REQUEST, bytes([PacketID.POSITION.value]))
    Packet(Packet ID: PacketID.REQUEST, Device ID: DeviceID.LINEAR_JAWS, Data: b'\x03')
"""

from __future__ import annotations

import struct

from cobs import cobs
from crc import Calculator, Configuration

from pybravo.protocol.device_id import DeviceID
from pybravo.protocol.packet_id import PacketID


class Packet:
    """A serial packet defined using the Reach serial specification."""

    _crc_calculator = Calculator(
        Configuration(
            width=8,
            polynomial=0x4D,
            init_value=0x00,
            final_xor_value=0xFF,
            reverse_input=True,
            reverse_output=True,
        )
    )

    def __init__(self, device_id: DeviceID, packet_id: PacketID, data: bytes) -> None:
        """Create a new serial packet.

        Args:
            device_id: The device ID that the packet is targeting.
            packet_id: The ID of the packet.
            data: The packet data.
        """
        self.device_id = device_id
        self.packet_id = packet_id
        self.data = data

    def __str__(self) -> str:
        """Print the packet as a string.

        Returns:
            A string description of the packet.
        """
        return (
            f"Packet(Packet ID: {self.packet_id}, Device ID: {self.device_id},"
            f" Data: {self.data!r})"
        )

    def encode(self) -> bytes:
        """Encode the serial data using the COBS encoding algorithm.

        Returns:
            The encoded serial data.
        """
        data = self.data
        data += struct.pack(
            ">BBB",
            self.packet_id.value,
            self.device_id.value,
            len(self.data) + 4,
        )
        data += struct.pack(">B", self._crc_calculator.checksum(data))

        return cobs.encode(data) + b"\x00"

    @classmethod
    def decode(cls, data: bytes) -> Packet:
        """Decode the provided serial data.

        Args:
            data: The encoded serial data to decode.

        Raises:
            ValueError: Invalid CRC value
            ValueError: The actual payload is not equal to the specified payload
            ValueError: The provided data is empty

        Returns:
            A packet with decoded serial data.
        """
        if len(data) <= 0:
            raise ValueError("Cannot decode an empty byte array!")

        if data[-1] == 0:
            decoded = bytearray(cobs.decode(data[:-1]))
        else:
            decoded = bytearray(cobs.decode(data))

        actual_crc = decoded.pop()

        if not cls._crc_calculator.verify(decoded, actual_crc):
            raise ValueError("The expected and actual CRC values do not match.")

        length = decoded.pop()

        if len(decoded) + 2 != length:
            raise ValueError(
                "The specified payload size is not equal to the actual payload size."
            )

        device_id = decoded.pop()
        packet_id = decoded.pop()

        return Packet(DeviceID(device_id), PacketID(packet_id), bytes(decoded))



# Copyright (c) 2023 Evan Palmer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest  # noqa

from pybravo.protocol import DeviceID, Packet, PacketID  # noqa


def test_packet_encoding() -> None:
    """Test that the packet encoding properly encodes serial data."""
    expected_encoding = bytes([0x06, 0x03, 0x60, 0x01, 0x05, 0x52, 0x00])

    packet = Packet(
        DeviceID.LINEAR_JAWS, PacketID.REQUEST, bytes([PacketID.POSITION.value])
    )

    assert expected_encoding == packet.encode()


def test_data_decoding() -> None:
    """Test that the serial data decoding properly decodes the serial data."""
    encoded_data = bytes([0x09, 0x01, 0x02, 0x03, 0x04, 0x01, 0xFF, 0x08, 0x5D, 0x00])
    decoded_data = bytes([0x01, 0x02, 0x03, 0x04])

    assert decoded_data == Packet.decode(encoded_data).data
